# Generated by Django 2.2.6 on 2019-11-13 13:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("katka", "0026_scmsteprun_step_type"),
    ]

    operations = [
        migrations.AddField(
            model_name="scmpipelinerun",
            name="first_parent_hash",
            field=models.CharField(
                blank=True,
                help_text="Commit hash of first parent commit, to determine order of commits. First commit has none.",
                max_length=64,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="scmpipelinerun",
            name="status",
            field=models.CharField(
                choices=[
                    ("initializing", "initializing"),
                    ("queued", "queued"),
                    ("in progress", "in progress"),
                    ("failed", "failed"),
                    ("success", "success"),
                    ("skipped", "skipped"),
                ],
                default="initializing",
                max_length=30,
            ),
        ),
    ]



from katka.auth import AuthType, has_full_access_scope
from katka.fields import username_on_model
from rest_framework import mixins, status
from rest_framework.exceptions import PermissionDenied
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet

from .permissions import HasFullScope, IsGroupAuthenticated


class UserOrScopeViewSet(GenericViewSet):
    permission_classes = [IsGroupAuthenticated | HasFullScope]

    def initialize_request(self, request, *args, **kwargs):
        drf_request = super().initialize_request(request, *args, **kwargs)

        auth_type = AuthType.ANONYMOUS
        if getattr(drf_request, "user", None) is not None and not drf_request.user.is_anonymous:
            auth_type = AuthType.GROUPS
        elif getattr(drf_request, "scopes", None) is not None:
            auth_type = AuthType.SCOPES

        # set it on the django HttpRequest
        request.katka_auth_type = auth_type

        return drf_request

    def get_queryset(self):
        # do not call super().get_queryset() since it raises NotImplementedError
        queryset = self.model.objects.all()
        if self.request.katka_auth_type is AuthType.SCOPES:
            if has_full_access_scope(self.request):
                return queryset

            raise PermissionDenied("Missing full access scope")

        return self.get_user_restricted_queryset(queryset)

    def get_user_restricted_queryset(self, queryset):
        raise NotImplementedError  # noqa


class ReadOnlyAuditMixin(mixins.RetrieveModelMixin, mixins.ListModelMixin, UserOrScopeViewSet):
    model = None

    def get_queryset(self):
        return super().get_queryset().exclude(deleted=True)


class UpdateAuditMixin(mixins.UpdateModelMixin, UserOrScopeViewSet):
    model = None

    def update(self, request, *args, **kwargs):
        with username_on_model(self.model, request.user.username):
            return super().update(request, *args, **kwargs)


class CreateAuditMixin(mixins.CreateModelMixin, UserOrScopeViewSet):
    def create(self, request, *args, **kwargs):
        with username_on_model(self.model, request.user.username):
            return super().create(request, *args, **kwargs)


class DestroyAuditMixin(mixins.DestroyModelMixin, UserOrScopeViewSet):
    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        instance.deleted = True

        with username_on_model(self.model, request.user.username):
            instance.save()

        return Response(status=status.HTTP_204_NO_CONTENT)


class AuditViewSet(CreateAuditMixin, UpdateAuditMixin, DestroyAuditMixin, ReadOnlyAuditMixin):
    pass


class FilterViewMixin:
    parameter_lookup_map = None

    """
    Uses the Serializer fields to construct GET Parameter filtering
    """

    def get_queryset(self):
        queryset = super().get_queryset()

        # Allow filtering on any field in serializer
        all_fields = self.model._meta.get_fields()
        filter_fields_lookup = {field.name: field.name for field in all_fields}
        filter_fields_keys = list(field.name for field in all_fields)

        # Also support a mapping from query parameter to django query field
        if self.parameter_lookup_map:
            filter_fields_keys += self.parameter_lookup_map.keys()
            filter_fields_lookup.update(self.parameter_lookup_map)

        # Loop through the keys to maintain proper order
        filters = {}
        for query_param in filter_fields_keys:
            django_lookup_field = filter_fields_lookup[query_param]
            value = self.request.query_params.get(query_param, None)
            if value is not None:
                filters[django_lookup_field] = value

        if filters:
            # Fetch distinct for all model fields, to prevent duplications due to SQL Joins
            queryset = queryset.distinct().filter(**filters)

        return queryset



# Generated by Django 2.1.7 on 2019-02-22 03:39

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("katka", "0009_scmsteprun"),
    ]

    operations = [
        migrations.RenameField(model_name="scmservice", old_name="type", new_name="scm_service_type",),
    ]



# Status is not in the model to allow usage in different models
PIPELINE_STATUS_INITIALIZING = "initializing"
PIPELINE_STATUS_QUEUED = "queued"
PIPELINE_STATUS_IN_PROGRESS = "in progress"
PIPELINE_STATUS_FAILED = "failed"
PIPELINE_STATUS_SUCCESS = "success"
PIPELINE_STATUS_SKIPPED = "skipped"

PIPELINE_STATUS_CHOICES = (
    (PIPELINE_STATUS_INITIALIZING, PIPELINE_STATUS_INITIALIZING),
    (PIPELINE_STATUS_QUEUED, PIPELINE_STATUS_QUEUED),
    (PIPELINE_STATUS_IN_PROGRESS, PIPELINE_STATUS_IN_PROGRESS),
    (PIPELINE_STATUS_FAILED, PIPELINE_STATUS_FAILED),
    (PIPELINE_STATUS_SUCCESS, PIPELINE_STATUS_SUCCESS),
    (PIPELINE_STATUS_SKIPPED, PIPELINE_STATUS_SKIPPED),
)

PIPELINE_FINAL_STATUSES = (
    PIPELINE_STATUS_SUCCESS,
    PIPELINE_STATUS_FAILED,
    PIPELINE_STATUS_SKIPPED,
)

STEP_STATUS_NOT_STARTED = "not started"
STEP_STATUS_IN_PROGRESS = "in progress"
STEP_STATUS_WAITING = "waiting"  # A step cannot proceed until some action is completed
STEP_STATUS_SKIPPED = "skipped"  # A step is skipped due to some conditional
STEP_STATUS_FAILED = "failed"
STEP_STATUS_SUCCESS = "success"

STEP_STATUS_CHOICES = (
    (STEP_STATUS_NOT_STARTED, STEP_STATUS_NOT_STARTED),
    (STEP_STATUS_IN_PROGRESS, STEP_STATUS_IN_PROGRESS),
    (STEP_STATUS_WAITING, STEP_STATUS_WAITING),
    (STEP_STATUS_SKIPPED, STEP_STATUS_SKIPPED),
    (STEP_STATUS_FAILED, STEP_STATUS_FAILED),
    (STEP_STATUS_SUCCESS, STEP_STATUS_SUCCESS),
)

STEP_FINAL_STATUSES = (STEP_STATUS_SKIPPED, STEP_STATUS_FAILED, STEP_STATUS_SUCCESS)
STEP_EXECUTED_STATUSES = (STEP_STATUS_FAILED, STEP_STATUS_SUCCESS)


RELEASE_STATUS_IN_PROGRESS = "in progress"
RELEASE_STATUS_FAILED = "failed"
RELEASE_STATUS_SUCCESS = "success"

RELEASE_STATUS_CHOICES = (
    (RELEASE_STATUS_IN_PROGRESS, RELEASE_STATUS_IN_PROGRESS),
    (RELEASE_STATUS_FAILED, RELEASE_STATUS_FAILED),
    (RELEASE_STATUS_SUCCESS, RELEASE_STATUS_SUCCESS),
)

TAG_PRODUCTION_CHANGE_STARTED = "production_change_start"
TAG_PRODUCTION_CHANGE_ENDED = "production_change_end"



from katka.auth import AuthType, has_full_access_scope
from rest_framework.permissions import BasePermission, IsAuthenticated


class IsGroupAuthenticated(IsAuthenticated):
    """
    Only if the request has a non anonymous User object and if that user is authenticated, allow
    """

    def has_permission(self, request, view):
        if request.katka_auth_type is not AuthType.GROUPS:
            return False

        return super().has_permission(request, view)


class HasFullScope(BasePermission):
    """
    Only if the request has scopes and if the full_access scope is present, allow
    """

    def has_permission(self, request, view):
        return has_full_access_scope(request)



# Generated by Django 2.2.7 on 2019-11-05 10:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("katka", "0024_scmsteprun_started_ended"),
    ]

    operations = [
        migrations.RenameField(model_name="application", old_name="created", new_name="created_at",),
        migrations.RenameField(model_name="application", old_name="modified", new_name="modified_at",),
        migrations.RenameField(model_name="applicationmetadata", old_name="created", new_name="created_at",),
        migrations.RenameField(model_name="applicationmetadata", old_name="modified", new_name="modified_at",),
        migrations.RenameField(model_name="credential", old_name="created", new_name="created_at",),
        migrations.RenameField(model_name="credential", old_name="modified", new_name="modified_at",),
        migrations.RenameField(model_name="credentialsecret", old_name="created", new_name="created_at",),
        migrations.RenameField(model_name="credentialsecret", old_name="modified", new_name="modified_at",),
        migrations.RenameField(model_name="project", old_name="created", new_name="created_at",),
        migrations.RenameField(model_name="project", old_name="modified", new_name="modified_at",),
        migrations.RenameField(model_name="scmpipelinerun", old_name="created", new_name="created_at",),
        migrations.RenameField(model_name="scmpipelinerun", old_name="modified", new_name="modified_at",),
        migrations.RenameField(model_name="scmrelease", old_name="created", new_name="created_at",),
        migrations.RenameField(model_name="scmrelease", old_name="ended", new_name="ended_at",),
        migrations.RenameField(model_name="scmrelease", old_name="modified", new_name="modified_at",),
        migrations.RenameField(model_name="scmrelease", old_name="started", new_name="started_at",),
        migrations.RenameField(model_name="scmrepository", old_name="created", new_name="created_at",),
        migrations.RenameField(model_name="scmrepository", old_name="modified", new_name="modified_at",),
        migrations.RenameField(model_name="scmservice", old_name="created", new_name="created_at",),
        migrations.RenameField(model_name="scmservice", old_name="modified", new_name="modified_at",),
        migrations.RenameField(model_name="scmsteprun", old_name="created", new_name="created_at",),
        migrations.RenameField(model_name="scmsteprun", old_name="ended", new_name="ended_at",),
        migrations.RenameField(model_name="scmsteprun", old_name="modified", new_name="modified_at",),
        migrations.RenameField(model_name="team", old_name="created", new_name="created_at",),
        migrations.RenameField(model_name="team", old_name="modified", new_name="modified_at",),
        migrations.RenameField(model_name="scmsteprun", old_name="started", new_name="started_at",),
    ]



from django.contrib import admin
from django.urls import include, path

from rest_framework import routers
from rest_framework_nested.routers import NestedSimpleRouter

from . import views

app_name = "katka_core"

router = routers.SimpleRouter()
router.register("teams", views.TeamViewSet, basename="teams")
router.register("projects", views.ProjectViewSet, basename="projects")
router.register("applications", views.ApplicationViewSet, basename="applications")
router.register("credentials", views.CredentialViewSet, basename="credentials")
router.register("scm-services", views.SCMServiceViewSet, basename="scm-services")
router.register("scm-repositories", views.SCMRepositoryViewSet, basename="scm-repositories")
router.register("scm-pipeline-runs", views.SCMPipelineRunViewSet, basename="scm-pipeline-runs")
router.register("scm-step-runs", views.SCMStepRunViewSet, basename="scm-step-runs")
router.register("scm-releases", views.SCMReleaseViewSet, basename="scm-releases")
router.register("update-scm-step-run", views.SCMStepRunUpdateStatusView, basename="update-scm-step-run")

secrets_router = NestedSimpleRouter(router, "credentials", lookup="credentials")
secrets_router.register("secrets", views.CredentialSecretsViewSet, basename="secrets")

app_metadata_router = NestedSimpleRouter(router, "applications", lookup="applications")
app_metadata_router.register("metadata", views.ApplicationMetadataViewSet, basename="metadata")

urlpatterns = [
    path("admin/", admin.site.urls),
    path("", include(router.urls)),
    path("", include(secrets_router.urls)),
    path("", include(app_metadata_router.urls)),
]



# Generated by Django 2.1.5 on 2019-02-12 02:12

import django.db.models.deletion
from django.db import migrations, models

import encrypted_model_fields.fields
import katka.fields


class Migration(migrations.Migration):

    dependencies = [
        ("katka", "0003_credential_and_unique_slugs"),
    ]

    operations = [
        migrations.CreateModel(
            name="CredentialSecret",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("created_username", katka.fields.AutoUsernameField(max_length=50)),
                ("modified", models.DateTimeField(auto_now=True)),
                ("modified_username", katka.fields.AutoUsernameField(max_length=50)),
                (
                    "status",
                    models.CharField(
                        choices=[("active", "active"), ("inactive", "inactive")], default="active", max_length=50
                    ),
                ),
                ("key", models.CharField(max_length=50)),
                ("value", encrypted_model_fields.fields.EncryptedCharField()),
                ("credential", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="katka.Credential")),
            ],
        ),
        migrations.AlterUniqueTogether(name="credentialsecret", unique_together={("credential", "key")},),
    ]



# Generated by Django 2.2.6 on 2019-10-21 07:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("katka", "0021_added_step_tags"),
    ]

    operations = [
        migrations.AlterField(
            model_name="scmrelease",
            name="status",
            field=models.CharField(
                choices=[("in progress", "in progress"), ("failed", "failed"), ("success", "success")],
                default="in progress",
                max_length=30,
            ),
        ),
    ]



# Generated by Django 2.1.5 on 2019-02-14 08:04

import uuid

from django.db import migrations, models

import katka.fields


class Migration(migrations.Migration):

    dependencies = [
        ("katka", "0004_credential_secret"),
    ]

    operations = [
        migrations.CreateModel(
            name="SCMService",
            fields=[
                ("created", models.DateTimeField(auto_now_add=True)),
                ("created_username", katka.fields.AutoUsernameField(max_length=50)),
                ("modified", models.DateTimeField(auto_now=True)),
                ("modified_username", katka.fields.AutoUsernameField(max_length=50)),
                (
                    "status",
                    models.CharField(
                        choices=[("active", "active"), ("inactive", "inactive")], default="active", max_length=50
                    ),
                ),
                (
                    "public_identifier",
                    models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False),
                ),
                ("type", models.CharField(max_length=48)),
                ("server_url", models.URLField()),
            ],
            options={"abstract": False,},
        ),
    ]



# Generated by Django 2.1.5 on 2019-02-19 03:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("katka", "0006_application_and_scm_repository"),
    ]

    operations = [
        migrations.RemoveField(model_name="application", name="status",),
        migrations.RemoveField(model_name="credential", name="status",),
        migrations.RemoveField(model_name="credentialsecret", name="status",),
        migrations.RemoveField(model_name="project", name="status",),
        migrations.RemoveField(model_name="scmrepository", name="status",),
        migrations.RemoveField(model_name="scmservice", name="status",),
        migrations.RemoveField(model_name="team", name="status",),
        migrations.AddField(model_name="application", name="deleted", field=models.BooleanField(default=False),),
        migrations.AddField(model_name="credential", name="deleted", field=models.BooleanField(default=False),),
        migrations.AddField(model_name="credentialsecret", name="deleted", field=models.BooleanField(default=False),),
        migrations.AddField(model_name="project", name="deleted", field=models.BooleanField(default=False),),
        migrations.AddField(model_name="scmrepository", name="deleted", field=models.BooleanField(default=False),),
        migrations.AddField(model_name="scmservice", name="deleted", field=models.BooleanField(default=False),),
        migrations.AddField(model_name="team", name="deleted", field=models.BooleanField(default=False),),
    ]



# Generated by Django 2.2.6 on 2019-10-24 08:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("katka", "0022_change_release_status_fields"),
    ]

    operations = [
        migrations.RemoveField(model_name="scmrelease", name="from_hash",),
        migrations.RemoveField(model_name="scmrelease", name="released",),
        migrations.RemoveField(model_name="scmrelease", name="to_hash",),
        migrations.AddField(model_name="scmrelease", name="ended", field=models.DateTimeField(null=True),),
        migrations.AddField(model_name="scmrelease", name="started", field=models.DateTimeField(null=True),),
    ]



# Generated by Django 2.1.7 on 2019-02-19 07:13

import uuid

import django.db.models.deletion
from django.db import migrations, models

import katka.fields


class Migration(migrations.Migration):

    dependencies = [
        ("katka", "0007_remove_status_add_deleted"),
    ]

    operations = [
        migrations.CreateModel(
            name="SCMPipelineRun",
            fields=[
                ("created", models.DateTimeField(auto_now_add=True)),
                ("created_username", katka.fields.AutoUsernameField(max_length=50)),
                ("modified", models.DateTimeField(auto_now=True)),
                ("modified_username", katka.fields.AutoUsernameField(max_length=50)),
                ("deleted", models.BooleanField(default=False)),
                (
                    "public_identifier",
                    models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False),
                ),
                ("commit_hash", models.CharField(max_length=64)),
                (
                    "status",
                    models.CharField(
                        choices=[("failed", "failed"), ("success", "success"), ("in progress", "in progress")],
                        default="in progress",
                        max_length=30,
                    ),
                ),
                ("steps_total", models.PositiveSmallIntegerField()),
                ("steps_completed", models.PositiveSmallIntegerField(default=0)),
                ("pipeline_yaml", models.TextField()),
                ("application", models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to="katka.Application")),
            ],
            options={"abstract": False,},
        ),
    ]



# Generated by Django 2.1.5 on 2019-05-07 01:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("katka", "0018_auto_20190506_0425"),
    ]

    operations = [
        migrations.AddField(
            model_name="scmsteprun", name="sequence_id", field=models.CharField(blank=True, max_length=30, null=True),
        ),
    ]



from django.contrib.auth.models import Group

from katka.auth import has_full_access_scope
from katka.constants import STEP_STATUS_CHOICES
from katka.models import (
    Application,
    ApplicationMetadata,
    Credential,
    CredentialSecret,
    Project,
    SCMPipelineRun,
    SCMRelease,
    SCMRepository,
    SCMService,
    SCMStepRun,
    Team,
)
from katka.serializer_fields import (
    ApplicationRelatedField,
    CredentialRelatedField,
    GroupNameField,
    ProjectRelatedField,
    SCMPipelineRunRelatedField,
    SCMRepositoryRelatedField,
    SCMServiceRelatedField,
    TeamRelatedField,
)
from rest_framework import serializers
from rest_framework.exceptions import PermissionDenied


class KatkaSerializer(serializers.ModelSerializer):
    pass


class TeamSerializer(KatkaSerializer):
    group = GroupNameField(queryset=Group.objects.all())

    class Meta:
        model = Team
        fields = ("public_identifier", "slug", "name", "group")

    def validate_group(self, group):
        if has_full_access_scope(self.context["request"]):
            querystring = Group.objects
        else:
            querystring = self.context["request"].user.groups

        querystring = querystring.filter(name=group.name)
        if not querystring.exists():
            raise PermissionDenied("Group does not exist or user is not a member of this group")

        return group


class ProjectSerializer(KatkaSerializer):
    team = TeamRelatedField()

    class Meta:
        model = Project
        fields = ("public_identifier", "slug", "name", "team")


class ApplicationSerializer(KatkaSerializer):
    project = ProjectRelatedField()
    scm_repository = SCMRepositoryRelatedField()

    class Meta:
        model = Application
        fields = ("public_identifier", "slug", "name", "project", "scm_repository", "active")


class CredentialSerializer(KatkaSerializer):
    team = TeamRelatedField()

    class Meta:
        model = Credential
        fields = ("public_identifier", "name", "team")


class CredentialSecretSerializer(serializers.ModelSerializer):
    credential = CredentialRelatedField()

    class Meta:
        model = CredentialSecret
        fields = ("key", "value", "credential")

    def to_internal_value(self, data):
        """Automatically add credential pk based on url kwargs"""
        data["credential"] = self.context["view"].kwargs["credentials_pk"]
        return super().to_internal_value(data)


class SCMServiceSerializer(serializers.ModelSerializer):
    class Meta:
        model = SCMService
        fields = ("public_identifier", "scm_service_type", "server_url")


class SCMRepositorySerializer(KatkaSerializer):
    credential = CredentialRelatedField()
    scm_service = SCMServiceRelatedField()

    class Meta:
        model = SCMRepository
        fields = ("public_identifier", "organisation", "repository_name", "credential", "scm_service")


class SCMPipelineRunSerializer(KatkaSerializer):
    application = ApplicationRelatedField()

    class Meta:
        model = SCMPipelineRun
        fields = (
            "public_identifier",
            "commit_hash",
            "first_parent_hash",
            "status",
            "steps_total",
            "steps_completed",
            "pipeline_yaml",
            "application",
            "scmrelease_set",
        )
        read_only_fields = ("scmrelease_set",)


class SCMStepRunSerializer(KatkaSerializer):
    scm_pipeline_run = SCMPipelineRunRelatedField()

    class Meta:
        model = SCMStepRun
        fields = (
            "public_identifier",
            "step_type",
            "slug",
            "name",
            "stage",
            "status",
            "output",
            "sequence_id",
            "scm_pipeline_run",
            "tags",
            "started_at",
            "ended_at",
        )


class SCMStepRunUpdateSerializer(KatkaSerializer):
    # it seems redundant to declare this field here as it is declared in the model, but in this
    # context it's a required field and in the model it's optional, thus the duplication.
    status = serializers.ChoiceField(required=True, choices=STEP_STATUS_CHOICES)

    class Meta:
        model = SCMStepRun
        fields = ("public_identifier", "status", "ended_at")

    def __init__(self, *args, **kwargs):
        kwargs["partial"] = False  # otherwise DRF will allow empty status
        super().__init__(*args, **kwargs)


class SCMReleaseSerializer(KatkaSerializer):
    scm_pipeline_runs = SCMPipelineRunRelatedField(required=False, read_only=True, many=True)

    class Meta:
        model = SCMRelease
        fields = ("public_identifier", "name", "started_at", "ended_at", "scm_pipeline_runs", "status")
        read_only_fields = ("started_at", "ended_at", "scm_pipeline_runs")


class SCMReleaseCreateSerializer(KatkaSerializer):
    scm_pipeline_runs = SCMPipelineRunRelatedField(required=False, many=True)

    class Meta:
        model = SCMRelease
        fields = ("name", "scm_pipeline_runs", "status")


class ApplicationMetadataSerializer(serializers.ModelSerializer):
    application = ApplicationRelatedField()

    class Meta:
        model = ApplicationMetadata
        fields = ("key", "value", "application")

    def to_internal_value(self, data):
        """Automatically add application pk based on url kwargs"""
        data["application"] = self.context["view"].kwargs["applications_pk"]
        return super().to_internal_value(data)



# Generated by Django 2.2.7 on 2019-11-20 09:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("katka", "0028_unique_commits"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="scmpipelinerun",
            options={
                "ordering": ["-created_at"],
                "verbose_name": "SCM pipeline",
                "verbose_name_plural": "SCM pipelines",
            },
        ),
        migrations.AlterModelOptions(
            name="scmrelease",
            options={"ordering": ["-created_at"], "verbose_name": "SCM release", "verbose_name_plural": "SCM releases"},
        ),
        migrations.AddIndex(
            model_name="scmpipelinerun",
            index=models.Index(fields=["-created_at"], name="katka_scmpi_created_7cc44e_idx"),
        ),
        migrations.AddIndex(
            model_name="scmrelease", index=models.Index(fields=["-created_at"], name="katka_scmre_created_b5704e_idx"),
        ),
    ]



import uuid

from django.contrib.auth.models import Group
from django.db import models

from encrypted_model_fields.fields import EncryptedCharField
from katka.auditedmodel import AuditedModel
from katka.constants import (
    PIPELINE_STATUS_CHOICES,
    PIPELINE_STATUS_INITIALIZING,
    RELEASE_STATUS_CHOICES,
    RELEASE_STATUS_IN_PROGRESS,
    STEP_STATUS_CHOICES,
    STEP_STATUS_NOT_STARTED,
)
from katka.fields import KatkaSlugField


class Team(AuditedModel):
    public_identifier = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    slug = KatkaSlugField(unique=True)
    name = models.CharField(max_length=100)
    group = models.ForeignKey(Group, on_delete=models.CASCADE)

    def __str__(self):  # pragma: no cover
        return f"{self.group.name}"


class Project(AuditedModel):
    public_identifier = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    slug = KatkaSlugField()
    name = models.CharField(max_length=100)
    team = models.ForeignKey(Team, on_delete=models.CASCADE)

    class Meta:
        unique_together = ("team", "slug")

    def __str__(self):  # pragma: no cover
        return f"{self.name}"


class Credential(AuditedModel):
    public_identifier = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=100)
    credential_type = models.CharField(max_length=50)
    team = models.ForeignKey(Team, on_delete=models.CASCADE)

    class Meta:
        unique_together = ("name", "team")

    def __str__(self):  # pragma: no cover
        return f"{self.credential_type}/{self.name}"


class CredentialSecret(AuditedModel):
    key = models.CharField(max_length=50)
    value = EncryptedCharField(max_length=200)
    credential = models.ForeignKey(Credential, on_delete=models.CASCADE)

    class Meta:
        unique_together = ("credential", "key")

    def __str__(self):  # pragma: no cover
        return f"{self.credential.name}/{self.key}"


class SCMService(AuditedModel):
    class Meta:
        verbose_name = "SCM service"
        verbose_name_plural = "SCM services"

    public_identifier = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    scm_service_type = models.CharField(max_length=48)
    server_url = models.URLField(null=False)

    def __str__(self):  # pragma: no cover
        return f"{self.scm_service_type}/{self.server_url}"


class SCMRepository(AuditedModel):
    class Meta:
        verbose_name = "SCM repository"
        verbose_name_plural = "SCM repositories"

    public_identifier = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    organisation = models.CharField(max_length=128)
    repository_name = models.CharField(max_length=128)
    credential = models.ForeignKey(Credential, on_delete=models.PROTECT)
    scm_service = models.ForeignKey(SCMService, on_delete=models.PROTECT)

    def __str__(self):  # pragma: no cover
        return f"{self.organisation}/{self.repository_name}"


class Application(AuditedModel):
    public_identifier = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    slug = KatkaSlugField()
    name = models.CharField(max_length=100)
    project = models.ForeignKey(Project, on_delete=models.PROTECT)
    scm_repository = models.OneToOneField(SCMRepository, on_delete=models.PROTECT)
    active = models.BooleanField(default=True)

    class Meta:
        unique_together = ("project", "slug")

    def __str__(self):  # pragma: no cover
        return f"{self.name}"


# Pipeline run results
class SCMPipelineRun(AuditedModel):
    class Meta:
        verbose_name = "SCM pipeline"
        verbose_name_plural = "SCM pipelines"
        constraints = (
            models.UniqueConstraint(fields=("commit_hash", "application"), name="unique commits per application"),
        )
        ordering = ["-created_at"]
        indexes = [models.Index(fields=["-created_at"])]

    public_identifier = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    commit_hash = models.CharField(max_length=64)  # A SHA-1 hash is 40 characters, SHA-256 is 64 characters
    first_parent_hash = models.CharField(
        max_length=64,
        help_text="Commit hash of first parent commit, to determine order of commits. First commit has none.",
        null=True,
        blank=True,
    )
    status = models.CharField(max_length=30, choices=PIPELINE_STATUS_CHOICES, default=PIPELINE_STATUS_INITIALIZING)
    steps_total = models.PositiveSmallIntegerField(default=0)
    steps_completed = models.PositiveSmallIntegerField(default=0)
    pipeline_yaml = models.TextField(default="---")
    application = models.ForeignKey(Application, on_delete=models.PROTECT)


class SCMStepRun(AuditedModel):
    class Meta:
        verbose_name = "SCM step"
        verbose_name_plural = "SCM steps"

    step_type = models.CharField(max_length=100, null=True)
    public_identifier = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    slug = KatkaSlugField(max_length=30)
    name = models.CharField(max_length=100)
    stage = models.CharField(max_length=100)
    status = models.CharField(max_length=30, choices=STEP_STATUS_CHOICES, default=STEP_STATUS_NOT_STARTED)
    output = models.TextField(blank=True)
    sequence_id = models.CharField(max_length=30, blank=True, null=True)
    # The format of a sequence ID is: <stage_nr>.<step_nr>-<parallel_nr>, with the following explanation:
    #
    #   stage_nr: the sequence number of the stage. The first stage gets number 1, the next 2, then 3, etc.
    #   step_nr: the sequence number of the step inside a stage. Each first step of the stage gets 1, then 2,
    #            then 3, etc.
    #   parallel_nr: these steps can be done in parallel so they only need a sequence so that the interface will show
    #                the steps in a consistent way.
    #
    # an example would be: "1.1-1" or "1.1" if there are no parallel steps. There should be zero padding, so if
    # there are more than 9 stages, it should be "01.1", or if there are more than 9 steps: "1.01".
    # This allows easy sorting for e.g. frontends.
    scm_pipeline_run = models.ForeignKey(SCMPipelineRun, on_delete=models.PROTECT)
    tags = models.TextField(blank=True)
    started_at = models.DateTimeField(blank=True, null=True)
    ended_at = models.DateTimeField(blank=True, null=True)


# SCM Releases, comprises a range of commits that are released
class SCMRelease(AuditedModel):
    class Meta:
        verbose_name = "SCM release"
        verbose_name_plural = "SCM releases"
        ordering = ["-created_at"]
        indexes = [models.Index(fields=["-created_at"])]

    public_identifier = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=255)
    status = models.CharField(max_length=30, choices=RELEASE_STATUS_CHOICES, default=RELEASE_STATUS_IN_PROGRESS)
    started_at = models.DateTimeField(null=True)
    ended_at = models.DateTimeField(null=True)
    scm_pipeline_runs = models.ManyToManyField(SCMPipelineRun)


class ApplicationMetadata(AuditedModel):
    key = models.CharField(max_length=50)
    value = models.CharField(max_length=200)
    application = models.ForeignKey(Application, on_delete=models.CASCADE)

    class Meta:
        unique_together = ("application", "key")

    def __str__(self):  # pragma: no cover
        return f"{self.application.name}/{self.key}"
